/*===================================================================================================
**
**	Author	:	Robert Bateman
**	E-Mail	:	rbateman@bournemouth.ac.uk
**	Brief	:	This Sourcefile is part of a series explaining how to load and render Alias Wavefront
**				Files somewhat efficently. If you are simkply after a reliable Obj Loader, then I would
**				Recommend version8; or possibly version9 and the supplied loader for extreme efficency.
**
**	Note	:	This Source Code is provided as is. No responsibility is accepted by myself for any
**				damage to hardware or software caused as a result of using this code. You are free to
**				make any alterations you see fit to this code for your own purposes, and distribute
**				that code either as part of a source code or binary executable package. All I ask is
**				for a little credit somewhere for my work!
** 
**				Any source improvements or bug fixes should be e-mailed to myself so I can update the
**				code for the greater good of the community at large. Credit will be given to the 
**				relevant people as always....
**				
**
**				Copyright (c) Robert Bateman, www.robthebloke.org, 2004
**
**				
**				National Centre for Computer Animation,
**				Bournemouth University,
**				Talbot Campus,
**				Bournemouth,
**				BH3 72F,
**				United Kingdom
**				ncca.bournemouth.ac.uk
**	
**
===================================================================================================*/




/*===================================================================================================
**
**	This is the first version of the objloader. At the moment all that we have is a simple linked 
**	list to allow us to store multiple loaded obj files. Over the coming versions, we will add more
**	to this source file to handle grater complexity of obj files until we have full support for 
**	polygonal and material data from the obj file.
**
===================================================================================================*/




/*===================================================================================================
**
**												Includes
**
**=================================================================================================*/

/*
**	Need the windows headers to be included for openGL when coding on WIN32
*/
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "ObjLoader.h"

/* Ignored until Version6 */
void ConvertMeshTo(ObjFile id,unsigned char What)
{
}
void SetLightPosition(float lx,float ly,float lz)
{
}
void SetTextures(ObjFile id,const char BumpTex[],const char BaseTex[])
{
}

/*===================================================================================================
**
**										Internal Data Structures
**
**=================================================================================================*/


/*
**	Each mesh is held as a structure with arrays of it's vertices, normals and texturing co-ordinates.
**	An Array of faces then references the arrays.
**
*/
typedef struct _ObjMesh
{
	ObjFile			 m_iMeshID;				/*	the ID of the mesh	*/
	struct _ObjMesh* m_pNext;				/*  pointer to the next Obj mesh */

} ObjMesh;



/*===================================================================================================
**
**										Global Variables
**
**=================================================================================================*/

/*
**	The global head of the linked list of meshes. This is a linked list because it is possible that you will be
**	loading and deleting meshes during the course of the programs execution.
*/
ObjMesh *g_LinkedListHead = NULL;

/* 
**	This is used to generate a unique ID for each Obj File
*/
unsigned int g_ObjIdGenerator=0;

/*
**	This function is only called from within the *.c file and is used to create an ObjMesh structure and
**	initialise its values (adds the mesh to the linked list also).
*/
ObjMesh *MakeOBJ( void )
{
	/*
	**	The pointer we will create the mesh at the end of
	*/
	ObjMesh *pMesh = NULL;

	/* allocate memory for the new obj mesh */
	pMesh = (ObjMesh*) malloc (sizeof(ObjMesh));

	/*	If the program asserts here, then there was a memory allocation failure	*/
	assert(pMesh);

	/*
	**	Initialise all the new mesh node
	*/
	pMesh->m_iMeshID			= ++g_ObjIdGenerator;
	pMesh->m_pNext				= g_LinkedListHead;
	g_LinkedListHead			= pMesh;
	return pMesh;
}

/*
**	This function loads an objfile and returns an ID number for it
*/
ObjFile LoadOBJ(const char *filename)
{

	/*
	**	Create the mesh structure and add it to the linked list
	*/
	ObjMesh *pMesh = MakeOBJ();

	/*
	**	Return the ID of the new mesh
	*/
	return pMesh->m_iMeshID;
}


/*
**	This function draws the specified obj file
*/
void DrawOBJ(ObjFile id)
{
	/*
	**	Because the meshes are on a linked list, we first need to find the
	**	mesh with the specified ID number so traverse the list.
	*/
	ObjMesh *pMesh = g_LinkedListHead;


	glDisable(GL_LIGHTING);


	/*
	**	Check to see if the mesh ID is valid.
	*/
	while(pMesh != NULL)
	{
		if(pMesh->m_iMeshID == id)
		{
			/*
			**	Found the mesh to draw, expect we have no data yet...
			*/
			return;
		}
		pMesh=pMesh->m_pNext;
	}
}

/*
**	Calling free() on NULL is VERY BAD in C, so make sure we
**	check all pointers before calling free.
*/
void DeleteMesh(ObjMesh* pMesh)
{
	/*
	**	If the pointer is valid
	*/
	if(pMesh)
	{
		/*	free the mesh */
		free( pMesh );
	}
}


/*
**	This function deletes the specified obj file
*/
void DeleteOBJ(ObjFile id)
{
	/*
	**	Create two pointers to walk through the linked list
	*/
	ObjMesh *pMesh = g_LinkedListHead;
	ObjMesh *pPrev=NULL;

	/*
	**	If we found the node that needs to be deleted
	*/
	while(pMesh != NULL)
	{
		if(pMesh->m_iMeshID == id)
		{
			if(pPrev) {
				pPrev->m_pNext = pMesh->m_pNext;
			}
			else {
				g_LinkedListHead = pMesh->m_pNext;
			}

			/*
			**	Free the memory allocated for this mesh
			*/
			DeleteMesh(pMesh);

			return;
		}

		pPrev = pMesh;
		pMesh = pMesh->m_pNext;
	}
}

/*
**	Delete all of the meshes starting from the front.
*/
void CleanUpOBJ(void)
{
	/* keep looping until all meshes are deleted */
	while(g_LinkedListHead) {
		DeleteOBJ(g_LinkedListHead->m_iMeshID);
	}
}

