/*===================================================================================================
**
**	Author	:	Robert Bateman
**	E-Mail	:	rbateman@bournemouth.ac.uk
**	Brief	:	This Sourcefile is part of a series explaining how to load and render Alias Wavefront
**				Files somewhat efficently. If you are simkply after a reliable Obj Loader, then I would
**				Recommend version8; or possibly version9 and the supplied loader for extreme efficency.
**
**	Note	:	This Source Code is provided as is. No responsibility is accepted by myself for any
**				damage to hardware or software caused as a result of using this code. You are free to
**				make any alterations you see fit to this code for your own purposes, and distribute
**				that code either as part of a source code or binary executable package. All I ask is
**				for a little credit somewhere for my work!
** 
**				Any source improvements or bug fixes should be e-mailed to myself so I can update the
**				code for the greater good of the community at large. Credit will be given to the 
**				relevant people as always....
**				
**
**				Copyright (c) Robert Bateman, www.robthebloke.org, 2004
**
**				
**				National Centre for Computer Animation,
**				Bournemouth University,
**				Talbot Campus,
**				Bournemouth,
**				BH3 72F,
**				United Kingdom
**				ncca.bournemouth.ac.uk
**	
**
===================================================================================================*/


/*===================================================================================================
**
**	This version now reads in and renders the normals and texturing co-ordinates. This loader will
**	only handle triangles at the moment, so polys with more than 3 edges will be rendered incorrectly.
**
===================================================================================================*/


/*===================================================================================================
**
**												Includes
**
**=================================================================================================*/

/*
**	Need the windows headers to be included for openGL when coding on WIN32
*/
#ifdef WIN32
#include <windows.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "ObjLoader.h"


/* Ignored until Version6 */
void ConvertMeshTo(ObjFile id,unsigned char What)
{
}
void SetLightPosition(float lx,float ly,float lz)
{
}
void SetTextures(ObjFile id,const char BumpTex[],const char BaseTex[])
{
}

/*===================================================================================================
**
**										Internal Data Structures
**
**=================================================================================================*/


/*
**	To start with, only deal with the obj file as a list of points. Essentually,
**	if we are able to draw the points out they should roughly describe the mesh.
**	So, lets look for the 'v' flag that says that the 3 floats after it describes a
**	Vertex. Essentually then, we need to have a structure for a vertex.
*/
typedef struct
{
	float x,			/*	The x component of the vertex position	*/
		  y,			/*	The y component of the vertex position	*/
		  z;			/*	The z component of the vertex position	*/
} ObjVertex;


/*
**	Now, the normal vector flag 'vn' could also do with a structure to represent it,
**	this will ultimately look very similar to the Vertex structure seeing as they have
**	similar data
*/
typedef struct
{
	float x,			/*	The x component of the normal vector	*/
		  y,			/*	The y component of the normal vector	*/
		  z;			/*	The z component of the normal vector	*/
} ObjNormal;


/*
**	A Texturing co-ordinate usually has only two values, the u and the v. Make a third
**	struct for this fact.
*/
typedef struct
{
	float u,			/*	The u parametric texturing co-ordinate	*/
		  v;			/*	The v parametric texturing co-ordinate	*/
} ObjTexCoord;


/*
**	Each face is composed of a number of `corners`. At each `corner` there will be at a
**	minimum, one vertex coordinate, and possibly a normal vector and a texturing co-ordinate.
**	Seeing as quite often a specific vertex is used by a number of faces, rather than store the
**	vertices in the face structure itself, we can use an integer reference to the position of the
**	vertex in the array. Basically it allows us the ability of storing the data as a 4byte value
**	rather than the 12 needed for a vertex.
**
**		Due to the fact that we are only going to deal with triangles on this occasion, we can use
**	static arrays within the structure for this. That makes our life somewhat easier because we don't
**	have to worry about re-sizing our face structure.
**
**	So our full face structure looks a bit like this :
*/
typedef struct
{
	unsigned int m_aVertexIndices[3],			/*	array of indicies that reference the vertex array in the mesh	*/
				 m_aNormalIndices[3],			/*	array of indicies that reference the normal array in the mesh	*/
				 m_aTexCoordIndicies[3];		/*	array of indicies that reference the uv coordinate array in the mesh	*/
} ObjFace;

/*
**	Each mesh is held as a structure with arrays of it's vertices, normals and texturing co-ordinates.
**	An Array of faces then references the arrays.
**
*/
typedef struct _ObjMesh
{
	ObjVertex		*m_aVertexArray;		/*	Array of vertices that make up this mesh	*/
	ObjNormal		*m_aNormalArray;		/*	Array of normals that make up this mesh		*/
	ObjTexCoord		*m_aTexCoordArray;		/*	Array of texturing co-ordinates that make up this mesh */
	ObjFace			*m_aFaces;				/*	Array of faces that make up this mesh */

	unsigned int	 m_iNumberOfVertices,	/*	The number of vertices in the m_aVertexArray array	*/
					 m_iNumberOfNormals,	/*	The number of normals in the m_aNormalArray array	*/
					 m_iNumberOfTexCoords,	/*	The number of uv's in the m_aTexCoordArray array	*/
					 m_iNumberOfFaces;		/*	The number of faces in the m_aFaces array			*/
	ObjFile			 m_iMeshID;				/*	the ID of the mesh	*/


	struct _ObjMesh* m_pNext;				/*	a pointer to the next mesh */ 

} ObjMesh;




/*===================================================================================================
**
**									Function Prototypes
**
**=================================================================================================*/

void		CleanUpOBJ(void);
void		DeleteOBJ(ObjFile id);
void		DeleteMesh(ObjMesh* pMesh);
ObjMesh*	MakeOBJ(void);
ObjFile		LoadOBJ(const char *filename);



/*===================================================================================================
**
**										Global Variables
**
**=================================================================================================*/

/*
**	Here I'm going to use an array of pointers to ObjMesh structures to load up my geometry. At the moment
**	MAX_OBJ is set to 10 so we can load up to a maximum of 10 objects at once. In the next part we'll extend
**	this to use a linked list to store as many as we want.
*/
ObjMesh *g_LinkedListHead=NULL;

/* 
**	This is used to generate a unique ID for each Obj File
*/
unsigned int g_ObjIdGenerator=0;

/*
**	This function is only called from within the *.c file and is used to create an ObjMesh structure and
**	initialise its values (adds the mesh to the linked list also).
*/
ObjMesh *MakeOBJ( void )
{
	/*
	**	The pointer we will create the mesh at the end of
	*/
	ObjMesh *pMesh = NULL;


	pMesh = (ObjMesh*) malloc (sizeof(ObjMesh));

	/*	If the program asserts here, then there was a memory allocation failure	*/
	assert(pMesh);

	/* generate a unique ID for the mesh */
	pMesh->m_iMeshID = ++g_ObjIdGenerator;


	/*
	**	Initialise all pointers to NULL
	*/
	pMesh->m_aFaces				= NULL;
	pMesh->m_aNormalArray		= NULL;
	pMesh->m_aTexCoordArray		= NULL;
	pMesh->m_aVertexArray		= NULL;
	pMesh->m_iNumberOfFaces		= 0;
	pMesh->m_iNumberOfNormals	= 0;
	pMesh->m_iNumberOfTexCoords = 0;
	pMesh->m_iNumberOfVertices	= 0;
	pMesh->m_pNext				= g_LinkedListHead;
	g_LinkedListHead			= pMesh;

	return pMesh;
}

ObjFile LoadOBJ(const char *filename)
{
	ObjMesh *pMesh=NULL;
	unsigned int vc=0,nc=0,tc=0,fc=0;
	char buffer[256];
	FILE *fp = NULL;

	/*
	**	Open the file for reading
	*/
	fp = fopen(filename,"r");

	/*
	**	If the program asserted here, then the file could not be found or opened
	*/
	if(!fp)
		return 0;

	/*
	**	Create the mesh structure and add it to the linked list
	*/
	pMesh = MakeOBJ();

	/*
	**	Run through the whole file looking for the various flags so that we can count
	**	up how many data elements there are. This is done so that we can make one memory
	**	allocation for the meshes data and then run through the file once more, this time
	**	reading in the data. It's purely done to reduce system overhead of memory allocation due
	**	to otherwise needing to reallocate data everytime we read in a new element.
	*/
	while(!feof(fp))
	{
		/*	Grab a line at a time	*/
		fgets(buffer,256,fp);

		/*	look for the 'vn' - vertex normal - flag	*/
		if( strncmp("vn ",buffer,3) == 0 )
		{
			++pMesh->m_iNumberOfNormals;
		}
		else

		/*	look for the 'vt' - texturing co-ordinate - flag  */
		if( strncmp("vt ",buffer,3) == 0 )
		{
			++pMesh->m_iNumberOfTexCoords;
		}
		else

		/*	look for the 'v ' - vertex co-ordinate - flag  */
		if( strncmp("v ",buffer,2) == 0 )
		{
			++pMesh->m_iNumberOfVertices;
		}
		else

		/*	look for the 'f ' - face - flag  */
		if( strncmp("f ",buffer,2) == 0 )
		{
			++pMesh->m_iNumberOfFaces;
		}
	}

	/*
	**	close the file for now....
	*/
	fclose(fp);

	/*
	**	Allocate the memory for the data arrays and check that it allocated ok
	*/
	pMesh->m_aVertexArray	= (ObjVertex*  )malloc( pMesh->m_iNumberOfVertices	* sizeof(ObjVertex)	  );
	assert(pMesh->m_aVertexArray);

	pMesh->m_aNormalArray	= (ObjNormal*  )malloc( pMesh->m_iNumberOfNormals	* sizeof(ObjNormal)	  );
	assert(pMesh->m_aNormalArray);

	pMesh->m_aTexCoordArray = (ObjTexCoord*)malloc( pMesh->m_iNumberOfTexCoords	* sizeof(ObjTexCoord) );
	assert(pMesh->m_aTexCoordArray);

	pMesh->m_aFaces			= (ObjFace*    )malloc( pMesh->m_iNumberOfFaces		* sizeof(ObjFace)	  );
	assert(pMesh->m_aFaces);

	/*
	**	Now we know how much data is contained in the file and we've allocated memory to hold it all.
	**	What we can therefore do now, is load up all of the data in the file easily.
	*/
	fp = fopen(filename,"r");

	while(!feof(fp))
	{
		/*	Grab a line at a time	*/
		fgets(buffer,256,fp);

		/*	look for the 'vn' - vertex normal - flag	*/
		if( strncmp("vn ",buffer,3) == 0 )
		{
			sscanf((buffer+2),"%f%f%f",
							&pMesh->m_aNormalArray[ nc ].x,
							&pMesh->m_aNormalArray[ nc ].y,
							&pMesh->m_aNormalArray[ nc ].z);
			++nc;
		}
		else

		/*	look for the 'vt' - texturing co-ordinate - flag  */
		if( strncmp("vt ",buffer,3) == 0 )
		{
			sscanf((buffer+2),"%f%f",
							&pMesh->m_aTexCoordArray[ tc ].u,
							&pMesh->m_aTexCoordArray[ tc ].v);
			++tc;
		}
		else

		/*	look for the 'v ' - vertex co-ordinate - flag  */
		if( strncmp("v ",buffer,2) == 0 )
		{
			sscanf((buffer+1),"%f%f%f",
							&pMesh->m_aVertexArray[ vc ].x,
							&pMesh->m_aVertexArray[ vc ].y,
							&pMesh->m_aVertexArray[ vc ].z);
			++vc;
		}
		else

		/*	look for the 'f ' - face - flag  */
		if( strncmp("f ",buffer,2) == 0 )
		{
			unsigned int i;

			/*
			**	Create a pointer to the face purely so that there is less to type !
			*/
			ObjFace *pf = &pMesh->m_aFaces[fc];

			/*
			**	Read in the nine indices that make up a face
			*/
			sscanf(buffer+2, "%d/%d/%d %d/%d/%d %d/%d/%d",
					&pf->m_aVertexIndices   [0],
					&pf->m_aTexCoordIndicies[0],
					&pf->m_aNormalIndices   [0],
					&pf->m_aVertexIndices   [1],
					&pf->m_aTexCoordIndicies[1],
					&pf->m_aNormalIndices   [1],
					&pf->m_aVertexIndices   [2],
					&pf->m_aTexCoordIndicies[2],
					&pf->m_aNormalIndices   [2] );

			/*
			**	need to reduce the indices by 1 because array indices start at 0, obj starts at 1
			*/
			for(i=0;i<3;i++)
			{
				--pf->m_aTexCoordIndicies[i];
				--pf->m_aNormalIndices   [i];
				--pf->m_aVertexIndices   [i];
			}

			/*
			**	Incremet the face count
			*/
			++fc;
		}
	}

	fclose(fp);

	return pMesh->m_iMeshID;
}




void DrawOBJ(ObjFile id)
{
	/*
	**	Because the meshes are on a linked list, we first need to find the
	**	mesh with the specified ID number so traverse the list.
	*/
	ObjMesh *pMesh = g_LinkedListHead;

	/*
	**	Check to see if the mesh ID is valid.
	*/
	while(pMesh != NULL)
	{
		if(pMesh->m_iMeshID == id)
		{
			/*
			**	In order to draw our mesh, we would basically like something of the following :
			**
			**	for_each_polygon_in_mesh
			**	{
			**		start_drawing_poly;
			**		for_each_vertex_in_the_poly
			**		{
			**			use_the_vertex_index_to_find_vertex_in_the_mesh's_array
			**		}
			**	}
			*/
			unsigned int i;
			glBegin(GL_TRIANGLES);
			for(i=0;i<pMesh->m_iNumberOfFaces;i++)
			{
				unsigned int j;
				ObjFace *pf = &pMesh->m_aFaces[i];

				for(j=0;j<3;j++)
				{
					glTexCoord2f( pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].u,
								pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].v);
					glNormal3f( pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].x,
								pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].y,
								pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].z);
					glVertex3f( pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].x,
								pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].y,
								pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].z);
				}
			}
			glEnd();

			return;
		}
		pMesh = pMesh->m_pNext;
	}
}

/*
**	Calling free() on NULL is VERY BAD in C, so make sure we
**	check all pointers before calling free.
*/
void DeleteMesh(ObjMesh* pMesh)
{
	/*
	**	If the pointer is valid
	*/
	if(pMesh)
	{
		/*	delete the face array */
		if(pMesh->m_aFaces)
		{
			free(pMesh->m_aFaces);
			pMesh->m_aFaces = NULL;
		}

		/*	delete the vertex array */
		if(pMesh->m_aVertexArray)
		{
			free(pMesh->m_aVertexArray);
			pMesh->m_aVertexArray = NULL;
		}

		/*	delete the normal array */
		if(pMesh->m_aNormalArray)
		{
			free(pMesh->m_aNormalArray);
			pMesh->m_aNormalArray = NULL;
		}

		/*	delete the texturing co-ordinate array */
		if(pMesh->m_aTexCoordArray)
		{
			free(pMesh->m_aTexCoordArray);
			pMesh->m_aTexCoordArray = NULL;
		}

		/*	free the mesh */
		free( pMesh );
	}
}

void DeleteOBJ(ObjFile id)
{
	/*
	**	Because the meshes are on a linked list, we first need to find the
	**	mesh with the specified ID number so traverse the list.
	*/
	ObjMesh *pMesh = g_LinkedListHead;
	ObjMesh *pPrev = NULL; 

	/*
	**	Check to see if the mesh ID is valid.
	*/
	while(pMesh != NULL)
	{
		if(pMesh->m_iMeshID == id)
		{
			if(!pPrev)
			{
				g_LinkedListHead = pMesh->m_pNext;
			}
			else {
				pPrev->m_pNext = pMesh->m_pNext;
			}

			/*
			**	Free the memory allocated for this mesh
			*/
			DeleteMesh(pMesh);

			return;
		}

		pPrev = pMesh;
		pMesh = pMesh->m_pNext;
	}
}

/*
**	Delete all of the meshes starting from the front.
*/
void CleanUpOBJ(void)
{
	while(g_LinkedListHead)
	{
		DeleteOBJ(g_LinkedListHead->m_iMeshID);
	}
}
