/*===================================================================================================
**
**	Author	:	Robert Bateman
**	E-Mail	:	rbateman@bournemouth.ac.uk
**	Brief	:	This Sourcefile is part of a series explaining how to load and render Alias Wavefront
**				Files somewhat efficently. If you are simkply after a reliable Obj Loader, then I would
**				Recommend version8; or possibly version9 and the supplied loader for extreme efficency.
**
**	Note	:	This Source Code is provided as is. No responsibility is accepted by myself for any
**				damage to hardware or software caused as a result of using this code. You are free to
**				make any alterations you see fit to this code for your own purposes, and distribute
**				that code either as part of a source code or binary executable package. All I ask is
**				for a little credit somewhere for my work!
** 
**				Any source improvements or bug fixes should be e-mailed to myself so I can update the
**				code for the greater good of the community at large. Credit will be given to the 
**				relevant people as always....
**				
**
**				Copyright (c) Robert Bateman, www.robthebloke.org, 2004
**
**				
**				National Centre for Computer Animation,
**				Bournemouth University,
**				Talbot Campus,
**				Bournemouth,
**				BH3 72F,
**				United Kingdom
**				ncca.bournemouth.ac.uk
**	
**
===================================================================================================*/




/*===================================================================================================
**
**	This version now adds support for display lists and vertex arrays. The loaded mesh data can now
**	be converted to vertex array or display lists to speed up rendering time.
**
===================================================================================================*/



/*===================================================================================================
**
**												Includes
**
**=================================================================================================*/

/*
**	Need the windows headers to be included for openGL when coding on WIN32
*/
#ifdef WIN32
#include <windows.h>
#endif

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "ObjLoader.h"


void SetLightPosition(float lx,float ly,float lz)
{
}
void SetTextures(ObjFile id,const char BumpTex[],const char BaseTex[])
{
}

/*===================================================================================================
**
**										Internal Data Structures
**
**=================================================================================================*/


/*
**	To start with, only deal with the obj file as a list of points. Essentually,
**	if we are able to draw the points out they should roughly describe the mesh.
**	So, lets look for the 'v' flag that says that the 3 floats after it describes a
**	Vertex. Essentually then, we need to have a structure for a vertex.
*/
typedef struct
{
	float x,			/*	The x component of the vertex position	*/
		  y,			/*	The y component of the vertex position	*/
		  z;			/*	The z component of the vertex position	*/
} ObjVertex;


/*
**	Now, the normal vector flag 'vn' could also do with a structure to represent it,
**	this will ultimately look very similar to the Vertex structure seeing as they have
**	similar data
*/
typedef struct
{
	float x,			/*	The x component of the normal vector	*/
		  y,			/*	The y component of the normal vector	*/
		  z;			/*	The z component of the normal vector	*/
} ObjNormal;


/*
**	A Texturing co-ordinate usually has only two values, the u and the v. Make a third
**	struct for this fact.
*/
typedef struct
{
	float u,			/*	The u parametric texturing co-ordinate	*/
		  v;			/*	The v parametric texturing co-ordinate	*/
} ObjTexCoord;


/*
**	Each face is composed of a number of 'corners'. At each 'corner' there will be at a
**	minimum, one vertex coordinate, and possibly a normal vector and a texturing co-ordinate.
**	Seeing as quite often a specific vertex is used by a number of faces, rather than store the
**	vertices in the face structure itself, we can use an integer reference to the position of the
**	vertex in the array. Basically it allows us the ability of storing the data as a 4byte value
**	rather than the 12 needed for a vertex.
**
**	To summaraise,
**
**		Each face is comprised of a number of 'corners', in this case the number will be held in
**	the value "m_iVertexCount". Taking into account the worst case scenario, for each corner, we
**	could have a vertex, normal and a texturing co-ordinate (usual actually). So we may need to hold
**	3 * m_iVertexCount number of indices for our faces. It kindof makes sense to basically dynamically
**	allocate an array to hold the vertex indicies, an array for the normal indices and one for the uv
**	coords. These are m_aVertexIndices, m_aNormalIndices, and m_aTexCoordIndicies respectivley.
**
**	So our full face structure looks a bit like this :
*/
typedef struct
{
	unsigned int  m_aVertexIndices[3],		/*	array of indicies that reference the vertex array in the mesh	*/
				  m_aNormalIndices[3],		/*	array of indicies that reference the normal array in the mesh	*/
				  m_aTexCoordIndicies[3];	/*	array of indicies that reference the uv coordinate array in the mesh	*/

} ObjFace;

/*
**	Each mesh is held as a structure with arrays of it's vertices, normals and texturing co-ordinates.
**	An Array of faces then references the arrays. There is also a pointer to the next node in the linked
**	list.
*/
typedef struct _ObjMesh
{
	ObjVertex		*m_aVertexArray;		/*	Array of vertices that make up this mesh	*/
	ObjNormal		*m_aNormalArray;		/*	Array of normals that make up this mesh		*/
	ObjTexCoord		*m_aTexCoordArray;		/*	Array of texturing co-ordinates that make up this mesh */
	ObjFace			*m_aFaces;				/*	Array of faces that make up this mesh */
	unsigned int	*m_aIndices;			/*	Array of triangle face indices  */

	unsigned int	 m_iNumberOfVertices,	/*	The number of vertices in the m_aVertexArray array	*/
					 m_iNumberOfNormals,	/*	The number of normals in the m_aNormalArray array	*/
					 m_iNumberOfTexCoords,	/*	The number of uv's in the m_aTexCoordArray array	*/
					 m_iNumberOfFaces;		/*	The number of faces in the m_aFaces array			*/

	struct _ObjMesh *m_pNext;				/*	Each mesh will be stored internally as a node on a linked list */
	ObjFile			 m_iMeshID;				/*	the ID of the mesh	*/
	unsigned char	 m_iMode;				/*	Display mode --> Either 0 - normal, 1 - display list or 2 - vertex array */
	unsigned int	 m_iDisplayListNum;		/*	Used only if the mode is set to display list - holds the display list number */

} ObjMesh;

/*===================================================================================================
**
**										Global Variables
**
**=================================================================================================*/

/*
**	The global head of the linked list of meshes. This is a linked list because it is possible that you will be
**	loading and deleting meshes during the course of the programs execution.
*/
ObjMesh *g_LinkedListHead = NULL;

/* 
**	This is used to generate a unique ID for each Obj File
*/
unsigned int g_ObjIdGenerator=0;

/*
**	This function will take a pointer to the mesh and the number of extra triangles that
**	are required. It will then resize the polygon face array within the mesh and return
**	the index of the first new face that was allocated.
*/
unsigned int ResizeTriangleArray(ObjMesh *pMesh,unsigned int iExtraTriangleCount)
{
	/*
	**	Create a temporary pointer to the old face array in the mesh
	*/
	ObjFace *pTempFaces = pMesh->m_aFaces;

	/*
	**	Keep track of the old face count
	*/
	unsigned int iTemp  = pMesh->m_iNumberOfFaces;

	/*
	**	Increase the number of faces that will be stored in the mesh
	*/
	pMesh->m_iNumberOfFaces += iExtraTriangleCount;

	/*
	**	Allocate the number of faces now required by this mesh. Note, you could use
	**	realloc for this, but it's just my personal preference to do it this way.
	*/
	pMesh->m_aFaces = (ObjFace*)malloc(pMesh->m_iNumberOfFaces*sizeof(ObjFace));

	/*
	**	If we assert here then the memory allocation for the meshes failed
	*/
	assert(pMesh->m_aFaces);

	/*
	**	Copy across the data that was stored in the old array
	*/
	memcpy(pMesh->m_aFaces,pTempFaces,iTemp*sizeof(ObjFace));

	/*
	**	Delete the data held in the previous array
	*/
	free(pTempFaces);

	memset(&pMesh->m_aFaces[ iTemp ],0,iExtraTriangleCount*sizeof(ObjFace) );

	/*
	**	return the old size of the array (thats the same as the index of the first new triangle).
	*/
	return iTemp;

}


/*
**	This function is only called from within the *.c file and is used to create an ObjMesh structure and
**	initialise its values (adds the mesh to the linked list also).
*/
ObjMesh *MakeOBJ( void )
{

	/*
	**	The pointer we will create the mesh at the end of
	*/
	ObjMesh *pMesh = NULL;


	pMesh = (ObjMesh*) malloc (sizeof(ObjMesh));

	/*	If the program asserts here, then there was a memory allocation failure	*/
	assert(pMesh);

	/*
	**	Initialise all pointers to NULL
	*/
	pMesh->m_aFaces				= NULL;
	pMesh->m_aNormalArray		= NULL;
	pMesh->m_aTexCoordArray		= NULL;
	pMesh->m_aVertexArray		= NULL;
	pMesh->m_aIndices			= NULL;
	pMesh->m_iNumberOfFaces		= 0;
	pMesh->m_iNumberOfNormals	= 0;
	pMesh->m_iNumberOfTexCoords = 0;
	pMesh->m_iNumberOfVertices	= 0;
	pMesh->m_iMode				= 0;
	pMesh->m_iMeshID			= ++g_ObjIdGenerator;


	/*
	**	Insert the mesh at the beginning of the linked list
	*/
	pMesh->m_pNext				= g_LinkedListHead;
	g_LinkedListHead			= pMesh;

	return pMesh;
}

void GetIndices(char *pString, unsigned int *pVert, unsigned int *pNorm, unsigned int *pUV)
{
	/*
	**	Use 3 pointers to hold references to the bits of the string where the indices start
	**	for the vertex, normal and texturing co-ordinate respectivly
	*/
	char *pV = pString,
		 *pN = NULL,
		 *pT = NULL;

	/*
	**	Walk through the string a character at a time looking for the '/' character. The first one
	**	we find will may have the uv - coord index behind it. The '/' character will be removed by
	**	setting it to NULL, If for example the texturing co-ord is missing, then you'd expect "//" to
	**	appear. This while loop would set both of those to NULL, that would give you a NULL string
	**	for the uv coord. This allows us to test the string later to see if there is or isn't a
	**	value there.
	*/
	while( *pString != '\0' )
	{
		if(*pString == '/')
		{
			*pString = '\0';
			if(!pT)
			{
				pT = pString+1;
			}
			else
			{
				pN = pString+1;
			}
		}
		pString++;
	}

	/*
	**	Read the vertex index
	*/
	sscanf(pV,"%d",pVert);

	/*
	**	If a normal is there, then read it's value, otherwise set the default of 1
	**	(1 is used because we will decrease the value by 1 later thus giving us 0)
	*/
	if( pN == NULL || *pN == '\0' )
	{
		*pNorm = 1;
	}
	else
	{
		sscanf(pN,"%d",pNorm);
	}

	/*
	**	If a texture co-ordinate is there, then read it's value, otherwise set the default of 1
	**	(1 is used because we will decrease the value by 1 later thus giving us 0)
	*/
	if( pT == NULL || *pT == '\0' )
	{
		*pUV = 1;
	}
	else
	{
		sscanf(pT,"%d",pUV);
	}
}

ObjFile LoadOBJ(const char *filename)
{
	ObjMesh *pMesh=NULL;
	unsigned int vc=0,nc=0,tc=0,fc=0;
	char buffer[256];
	FILE *fp = NULL;

	/*
	**	Open the file for reading
	*/
	fp = fopen(filename,"r");

	/*
	**	If the program asserted here, then the file could not be found or opened
	*/
	assert(fp);

	/*
	**	Create the mesh structure and add it to the linked list
	*/
	pMesh = MakeOBJ();

	/*
	**	Run through the whole file looking for the various flags so that we can count
	**	up how many data elements there are. This is done so that we can make one memory
	**	allocation for the meshes data and then run through the file once more, this time
	**	reading in the data. It's purely done to reduce system overhead of memory allocation due
	**	to otherwise needing to reallocate data everytime we read in a new element.
	*/
	while(!feof(fp))
	{
		memset(buffer,0,200);

		/*	Grab a line at a time	*/
		fgets(buffer,256,fp);

		/*	look for the 'vn' - vertex normal - flag	*/
		if( strncmp("vn ",buffer,3) == 0 )
		{
			++pMesh->m_iNumberOfNormals;
		}
		else

		/*	look for the 'vt' - texturing co-ordinate - flag  */
		if( strncmp("vt ",buffer,3) == 0 )
		{
			++pMesh->m_iNumberOfTexCoords;
		}
		else

		/*	look for the 'v ' - vertex co-ordinate - flag  */
		if( strncmp("v ",buffer,2) == 0 )
		{
			++pMesh->m_iNumberOfVertices;
		}
		else

		/*	look for the 'f ' - face - flag  */
		if( strncmp("f ",buffer,2) == 0 )
		{
			++pMesh->m_iNumberOfFaces;
		}
	}

	/*
	**	close the file for now....
	*/
	fclose(fp);

	/*
	**	Allocate the memory for the data arrays and check that it allocated ok
	*/
	pMesh->m_aVertexArray	= (ObjVertex*  )malloc( pMesh->m_iNumberOfVertices	* sizeof(ObjVertex)	  );
	assert(pMesh->m_aVertexArray);

	/*	there are occasionally times when the obj does not have any normals in it */
	if( pMesh->m_iNumberOfNormals > 0 )
	{
		pMesh->m_aNormalArray	= (ObjNormal*  )malloc( pMesh->m_iNumberOfNormals	* sizeof(ObjNormal)	  );
		assert(pMesh->m_aNormalArray);
	}

	/*	there are occasionally times when the obj does not have any tex coords in it */
	if( pMesh->m_iNumberOfTexCoords > 0 )
	{
		pMesh->m_aTexCoordArray = (ObjTexCoord*)malloc( pMesh->m_iNumberOfTexCoords	* sizeof(ObjTexCoord) );
		assert(pMesh->m_aTexCoordArray);
	}

	pMesh->m_aFaces			= (ObjFace*    )malloc( pMesh->m_iNumberOfFaces		* sizeof(ObjFace)	  );
	assert(pMesh->m_aFaces);

	/*
	**	Now we know how much data is contained in the file and we've allocated memory to hold it all.
	**	What we can therefore do now, is load up all of the data in the file easily.
	*/
	fp = fopen(filename,"r");

	while(!feof(fp))
	{
		memset(buffer,0,255);

		/*	Grab a line at a time	*/
		fgets(buffer,256,fp);

		/*	look for the 'vn' - vertex normal - flag	*/
		if( strncmp("vn ",buffer,3) == 0 )
		{
			sscanf((buffer+2),"%f%f%f",
							&pMesh->m_aNormalArray[ nc ].x,
							&pMesh->m_aNormalArray[ nc ].y,
							&pMesh->m_aNormalArray[ nc ].z);
			++nc;
		}
		else

		/*	look for the 'vt' - texturing co-ordinate - flag  */
		if( strncmp("vt ",buffer,3) == 0 )
		{
			sscanf((buffer+2),"%f%f",
							&pMesh->m_aTexCoordArray[ tc ].u,
							&pMesh->m_aTexCoordArray[ tc ].v);
			++tc;
		}
		else

		/*	look for the 'v ' - vertex co-ordinate - flag  */
		if( strncmp("v ",buffer,2) == 0 )
		{
			sscanf((buffer+1),"%f%f%f",
							&pMesh->m_aVertexArray[ vc ].x,
							&pMesh->m_aVertexArray[ vc ].y,
							&pMesh->m_aVertexArray[ vc ].z);
			++vc;
		}
		else

		/*	look for the 'f ' - face - flag  */
		if( strncmp("f ",buffer,2) == 0 )
		{
			/*
			**	some data for later....
			*/
			char *pSplitString = NULL;
			unsigned int i,ii = 0;
			unsigned int iNewTrianglesIndex = 0,
						 iNumberOfExtraTriangles = 0;
			unsigned int aFirstIndices[3],
						 aSecondIndices[3];

			ObjFace *pf = NULL;

			/*
			**	These next few lines are used to figure out how many '/' characters there
			**	are in the string. This gives us the information we need to find out how
			**	many vertices are used in this face (by dividing by two)
			*/
			for(i=0;i<strlen(buffer);i++)
			{
				if(buffer[i] == '/')
					ii++;
			}
			ii/=2;

			/*
			**	This is where it all gets confusing, What I'm going to do is to convert all polys
			**	With vertices above 3 to triangles. The number of triangles I require will be (ii-2).
			**	For example, with a poly that has vertices a,b,c,d,e; there will be 3 triangles comprised
			**	of verts (a,b,c), (a,c,d) and (a,d,e) respectively. We need to check the amount of triangles
			**	that will be needed for this polygon and resize the array accordingly.
			*/
			if( (ii-2) > 1 )
			{
				iNumberOfExtraTriangles = ii-3;
				iNewTrianglesIndex = ResizeTriangleArray(pMesh,iNumberOfExtraTriangles);
			}

			/*
			**	Pointer to the face we are currently dealing with. It's only used so that
			**	the code becomes more readable and I have less to type.
			*/
			pf = &pMesh->m_aFaces[ fc ];


			/*
			**	tokenise the string using strtok(). Basically this splits the string up
			**	and removes the spaces from each chunk. This way we only have to deal with
			**	one set of indices at a time for each of the poly's vertices.
			*/
			pSplitString = strtok((buffer+2)," \t\n");
			i=0;

			/*
			**	Get the vertex structures that we know WILL exists (there has to be a minimum of three)
			*/
			for(i=0;i<3;i++)
			{
				GetIndices( pSplitString,
							&pf->m_aVertexIndices[i],
							&pf->m_aNormalIndices[i],
							&pf->m_aTexCoordIndicies[i]);

				/* need to reduce the indices by 1 because array indices start at 0, obj starts at 1  */
				--pf->m_aTexCoordIndicies[i];
				--pf->m_aNormalIndices[i];
				--pf->m_aVertexIndices[i];

				/*
				**	In order to start constructing our triangles that are created after we split the poly,
				**	we need to store some of the indices that are going to be used for the extra additional
				**	triangles after this. Basically we store the first and last sets of indices from this
				**	first triangle becasue they will be used as the first & second sets for the next triangle.
				*/
				switch(i)
				{
				case 0:
					aFirstIndices[0] = pf->m_aVertexIndices[i];
					aFirstIndices[1] = pf->m_aNormalIndices[i];
					aFirstIndices[2] = pf->m_aTexCoordIndicies[i];
					break;

				case 2:
					aSecondIndices[0] = pf->m_aVertexIndices[i];
					aSecondIndices[1] = pf->m_aNormalIndices[i];
					aSecondIndices[2] = pf->m_aTexCoordIndicies[i];
					break;
				default:
					break;
				}

				pSplitString = strtok(NULL," \t\n");
			}
			++fc;

			/*
			**	Get the additional triangles that may be there
			*/
			for(i=0;i<iNumberOfExtraTriangles;i++)
			{
				pf = &pMesh->m_aFaces[ iNewTrianglesIndex + i ];

				GetIndices( pSplitString,
							&pf->m_aVertexIndices[2],
							&pf->m_aNormalIndices[2],
							&pf->m_aTexCoordIndicies[2]);

				/* need to reduce the indices by 1 because array indices start at 0, obj starts at 1  */
				--pf->m_aTexCoordIndicies[2];
				--pf->m_aNormalIndices[2];
				--pf->m_aVertexIndices[2];

				/*
				**	Now assign the other indices for the triangle
				*/
				pf->m_aVertexIndices[0]		= aFirstIndices[0];
				pf->m_aNormalIndices[0]		= aFirstIndices[1];
				pf->m_aTexCoordIndicies[0]	= aFirstIndices[2];

				pf->m_aVertexIndices[1]		= aSecondIndices[0];
				pf->m_aNormalIndices[1]		= aSecondIndices[1];
				pf->m_aTexCoordIndicies[1]	= aSecondIndices[2];

				/*
				**	Copy over the indices for the next triangle.
				*/
				aSecondIndices[0] = pf->m_aVertexIndices[2];
				aSecondIndices[1] = pf->m_aNormalIndices[2];
				aSecondIndices[2] = pf->m_aTexCoordIndicies[2];

				pSplitString = strtok(NULL," \t\n");
			}
		}
	}

	fclose(fp);

	return pMesh->m_iMeshID;
}




void DrawOBJ(ObjFile id)
{
	/*
	**	Because the meshes are on a linked list, we first need to find the
	**	mesh with the specified ID number so traverse the list.
	*/
	ObjMesh *pMesh = g_LinkedListHead;

	while(pMesh && pMesh->m_iMeshID != id)
	{
		pMesh = pMesh->m_pNext;
	}

	/*
	**	Check to see if the mesh ID is valid.
	*/
	if(pMesh != NULL)
	{
		/*
		**	Select which drawing mode to use
		*/
		switch(pMesh->m_iMode)
		{
		case 0:
			{
				/*
				**	In order to draw our mesh, we would basically like something of the following :
				**
				**	for_each_polygon_in_mesh
				**	{
				**		start_drawing_poly;
				**		for_each_vertex_in_the_poly
				**		{
				**			use_the_vertex_index_to_find_vertex_in_the_mesh's_array
				**		}
				**	}
				*/
				if( pMesh->m_aNormalArray   != NULL &&
					pMesh->m_aTexCoordArray != NULL )
				{
					unsigned int i;
					glBegin(GL_TRIANGLES);
					for(i=0;i<pMesh->m_iNumberOfFaces;i++)
					{
						unsigned int j;
						ObjFace *pf = &pMesh->m_aFaces[i];

						/*
						**	Draw the polygons with normals & uv co-ordinates
						*/
						for(j=0;j<3;j++)
						{
							glTexCoord2f( pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].u,
										  pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].v);
							glNormal3f( pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].x,
										pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].y,
										pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].z);
							glVertex3f( pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].x,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].y,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].z);
						}
					}
					glEnd();
				}
				else
				if( pMesh->m_aNormalArray   != NULL )
				{
					unsigned int i;
					glBegin(GL_TRIANGLES);
					for(i=0;i<pMesh->m_iNumberOfFaces;i++)
					{
						unsigned int j;
						ObjFace *pf = &pMesh->m_aFaces[i];

						/*
						**	Draw the polygons with Normals only
						*/
						for(j=0;j<3;j++)
						{
							glNormal3f( pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].x,
										pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].y,
										pMesh->m_aNormalArray[ pf->m_aNormalIndices[j] ].z);
							glVertex3f( pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].x,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].y,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].z);
						}
					}
					glEnd();
				}
				else
				if( pMesh->m_aTexCoordArray != NULL )
				{
					unsigned int i;
					glBegin(GL_TRIANGLES);
					for(i=0;i<pMesh->m_iNumberOfFaces;i++)
					{
						unsigned int j;
						ObjFace *pf = &pMesh->m_aFaces[i];

						/*
						**	Draw the polygons with Texturing co-ordinates only
						*/
						for(j=0;j<3;j++)
						{
							glTexCoord2f( pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].u,
										  pMesh->m_aTexCoordArray[ pf->m_aTexCoordIndicies[j] ].v);
							glVertex3f( pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].x,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].y,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].z);
						}
					}
					glEnd();
				}
				else
				{
					unsigned int i;
					glBegin(GL_TRIANGLES);
					for(i=0;i<pMesh->m_iNumberOfFaces;i++)
					{
						unsigned int j;
						ObjFace *pf = &pMesh->m_aFaces[i];

						/*
						**	Draw the polygons with Texturing co-ordinates only
						*/
						for(j=0;j<3;j++)
						{
							glVertex3f( pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].x,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].y,
										pMesh->m_aVertexArray[ pf->m_aVertexIndices[j] ].z);
						}
					}
					glEnd();
				}
			break;

			case DISPLAY_LIST:
				{
					glCallList(pMesh->m_iDisplayListNum);
				}
				break;

			case VERTEX_ARRAY:
				{
					/*
					**	Enable the vertex array
					*/
					glVertexPointer(3,GL_FLOAT,0,pMesh->m_aVertexArray);
					glEnableClientState(GL_VERTEX_ARRAY);

					/*
					**	Enable the normal array (if present)
					*/
					if(pMesh->m_aNormalArray)
					{
						glNormalPointer(GL_FLOAT,0,pMesh->m_aNormalArray);
						glEnableClientState(GL_NORMAL_ARRAY);
					}

					/*
					**	Enable the tex coord array (if present)
					*/
					if(pMesh->m_aTexCoordArray)
					{
						glTexCoordPointer(2,GL_FLOAT,0,pMesh->m_aTexCoordArray);
						glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					}

					/*
					**	Draw the vertex array data
					*/
				//	glDrawArrays(GL_TRIANGLES,0,pMesh->m_iNumberOfFaces);
					glDrawElements(GL_TRIANGLES,pMesh->m_iNumberOfFaces,GL_UNSIGNED_INT,pMesh->m_aIndices);

					/*
					**	Disable the arrays before exiting
					*/
					glDisableClientState(GL_VERTEX_ARRAY);
					glDisableClientState(GL_NORMAL_ARRAY);
					glDisableClientState(GL_TEXTURE_COORD_ARRAY);
				}
				break;

			default:
				break;
			}


		}
	}
}

/*
**	Calling free() on NULL is VERY BAD in C, so make sure we
**	check all pointers before calling free.
*/
void DeleteMesh(ObjMesh* pMesh)
{
	/*
	**	If the pointer is valid
	*/
	if(pMesh)
	{
		/*	delete the face array */
		if(pMesh->m_aFaces)
		{
			free(pMesh->m_aFaces);
			pMesh->m_aFaces = NULL;
		}

		/*	delete the vertex array */
		if(pMesh->m_aVertexArray)
		{
			free(pMesh->m_aVertexArray);
			pMesh->m_aVertexArray = NULL;
		}

		/*	delete the normal array */
		if(pMesh->m_aNormalArray)
		{
			free(pMesh->m_aNormalArray);
			pMesh->m_aNormalArray = NULL;
		}

		/*	delete the texturing co-ordinate array */
		if(pMesh->m_aTexCoordArray)
		{
			free(pMesh->m_aTexCoordArray);
			pMesh->m_aTexCoordArray = NULL;
		}

		/*	delete the texturing co-ordinate array */
		if(pMesh->m_aIndices)
		{
			free(pMesh->m_aIndices);
			pMesh->m_aIndices = NULL;
		}

		/*	delete the display list number	*/
		if(pMesh->m_iMode == DISPLAY_LIST)
		{
			glDeleteLists(pMesh->m_iDisplayListNum,1);
		}

		/*	free the mesh */
		free( pMesh );
	}
}

void DeleteOBJ(ObjFile id)
{
	/*
	**	Create two pointers to walk through the linked list
	*/
	ObjMesh *pCurr,
			*pPrev = NULL;

	/*
	**	Start traversing the list from the start
	*/
	pCurr = g_LinkedListHead;

	/*
	**	Walk through the list until we either reach the end, or
	**	we find the node we are looking for
	*/
	while(pCurr != NULL && pCurr->m_iMeshID != id)
	{
		pPrev = pCurr;
		pCurr = pCurr->m_pNext;
	}

	/*
	**	If we found the node that needs to be deleted
	*/
	if(pCurr != NULL)
	{
		/*
		**	If the pointer before it is NULL, then we need to
		**	remove the first node in the list
		*/
		if(pPrev == NULL)
		{
			g_LinkedListHead = pCurr->m_pNext;
		}

		/*
		**	Otherwise we are removing a node from somewhere else
		*/
		else
		{
			pPrev->m_pNext = pCurr->m_pNext;
		}

		/*
		**	Free the memory allocated for this mesh
		*/
		DeleteMesh(pCurr);
	}
}

/*
**	Delete all of the meshes starting from the front.
*/
void CleanUpOBJ(void)
{
	ObjMesh *pCurr;
	while(g_LinkedListHead != NULL)
	{
		pCurr = g_LinkedListHead;
		g_LinkedListHead = g_LinkedListHead->m_pNext;
		DeleteMesh(pCurr);
	}
}


void ConvertMeshTo(ObjFile id,unsigned char What)
{
	/*
	**	Because the meshes are on a linked list, we first need to find the
	**	mesh with the specified ID number so traverse the list.
	*/
	ObjMesh *pMesh = g_LinkedListHead;

	while(pMesh && pMesh->m_iMeshID != id)
	{
		pMesh = pMesh->m_pNext;
	}

	/*
	**	If the mesh has already been converted to a display list or a vertex array, return.
	*/
	if(pMesh->m_iMode != 0)
		return;

	switch(What)
	{
	case DISPLAY_LIST:
		{
			/*
			**	Generate a display list number
			*/
			pMesh->m_iDisplayListNum = glGenLists(1);

			/*
			**	Make the display list to draw the object
			*/
			glNewList(pMesh->m_iDisplayListNum,GL_COMPILE);
				DrawOBJ(id);
			glEndList();

			/*
			**	Assign the new mode value to the mesh
			*/
			pMesh->m_iMode = What;

			/*
			**	Delete all the data held on the node and set values to NULL
			*/
			if(pMesh->m_aFaces)
				free(pMesh->m_aFaces);
			pMesh->m_aFaces = NULL;

			if(pMesh->m_aVertexArray)
				free(pMesh->m_aVertexArray);
			pMesh->m_aVertexArray = NULL;

			if(pMesh->m_aNormalArray)
				free(pMesh->m_aNormalArray);
			pMesh->m_aNormalArray = NULL;

			if(pMesh->m_aTexCoordArray)
				free(pMesh->m_aTexCoordArray);
			pMesh->m_aTexCoordArray = NULL;

			/*
			**	Set counts to zero
			*/
			pMesh->m_iNumberOfFaces		= 0;
			pMesh->m_iNumberOfNormals	= 0;
			pMesh->m_iNumberOfTexCoords = 0;
			pMesh->m_iNumberOfVertices	= 0;

		}
		break;

	case VERTEX_BUFFER_OBJECT:
	case VERTEX_ARRAY:
		{
			unsigned int i,j,k,l,m=0;

			/*
			**	Create a temporary set of pointers to the old data....
			*/
			ObjVertex   *pVT = pMesh->m_aVertexArray;
			ObjNormal   *pNT = pMesh->m_aNormalArray;
			ObjTexCoord *pTT = pMesh->m_aTexCoordArray;

			/*
			**	This is used to keep track of the actual vertex array size created.
			**	Once the data has been organised into 
			*/
			unsigned int ActualVertexArraySize=0;

			/*
			**	Allocate a new set of arrays to hold the expanded data
			*/
			pMesh->m_aVertexArray		= (ObjVertex*)  malloc(pMesh->m_iNumberOfFaces*3*sizeof(ObjVertex));
			assert( pMesh->m_aVertexArray );
			if(pMesh->m_aNormalArray)
			{
				pMesh->m_aNormalArray   = (ObjNormal*)  malloc(pMesh->m_iNumberOfFaces*3*sizeof(ObjNormal));
				assert(pMesh->m_aNormalArray);
			}
			if(pMesh->m_aTexCoordArray)
			{
				pMesh->m_aTexCoordArray = (ObjTexCoord*)malloc(pMesh->m_iNumberOfFaces*3*sizeof(ObjTexCoord));
				assert(pMesh->m_aTexCoordArray);
			}


			/* 
			**	Allocate memory for the vertex array indices
			*/
			pMesh->m_aIndices	= (unsigned int*)malloc( pMesh->m_iNumberOfFaces*3*sizeof(unsigned int)	  );

			assert(pMesh->m_aVertexArray);

			/*
			**	write each vertex/normal/and uv co-ordinate out to the new array
			**	so that we can use vertex arrays....
			*/
			for(i=0,j=0;i<pMesh->m_iNumberOfFaces;i++)
			{
				for(k=0;k<3;k++,j++)
				{
					float v[3]={0},n[3]={0},t[2]={0};

					#define EPISILON 0.0001f
					#define FCOMP(A,B) (((A)>((B)-EPISILON)) && ((A)<((B)+EPISILON)))

					v[0] = pVT[ pMesh->m_aFaces[i].m_aVertexIndices[k] ].x;
					v[1] = pVT[ pMesh->m_aFaces[i].m_aVertexIndices[k] ].y;
					v[2] = pVT[ pMesh->m_aFaces[i].m_aVertexIndices[k] ].z;

					if(pMesh->m_aNormalArray)
					{
						n[0]   = pNT[ pMesh->m_aFaces[i].m_aNormalIndices[k] ].x;
						n[1]   = pNT[ pMesh->m_aFaces[i].m_aNormalIndices[k] ].y;
						n[2]   = pNT[ pMesh->m_aFaces[i].m_aNormalIndices[k] ].z;
					}

					if(pMesh->m_aTexCoordArray)
					{
						t[0]   = pTT[ pMesh->m_aFaces[i].m_aTexCoordIndicies[k] ].u;
						t[1]   = pTT[ pMesh->m_aFaces[i].m_aTexCoordIndicies[k] ].v;
					}
					
					for(l=0;l<ActualVertexArraySize;++l) 
					{
                        if( FCOMP( pMesh->m_aVertexArray[j].x, v[0] ) && 
							FCOMP( pMesh->m_aVertexArray[j].y, v[1] ) &&
							FCOMP( pMesh->m_aVertexArray[j].z, v[2] ) ) {

							if(pMesh->m_aNormalArray)
							{
								if( FCOMP( pMesh->m_aNormalArray[j].x, n[0] ) && 
									FCOMP( pMesh->m_aNormalArray[j].y, n[1] ) &&
									FCOMP( pMesh->m_aNormalArray[j].z, n[2] ) )
								{
									if(pMesh->m_aTexCoordArray)
									{
										if( FCOMP( pMesh->m_aTexCoordArray[j].u, t[0] ) && 
											FCOMP( pMesh->m_aTexCoordArray[j].v, t[1] )  )
										{
											goto reuse_element;
										}
									}
									else 
										goto reuse_element;
								}
							}
							else
							if(pMesh->m_aTexCoordArray)
							{
								if( FCOMP( pMesh->m_aTexCoordArray[j].u, t[0] ) && 
									FCOMP( pMesh->m_aTexCoordArray[j].v, t[1] )  )
								{
									goto reuse_element;
								}
							}
							else
								goto reuse_element;
                        }
					}
					pMesh->m_aVertexArray[j].x		 = v[0];
					pMesh->m_aVertexArray[j].y		 = v[1];
					pMesh->m_aVertexArray[j].z		 = v[2];

					if(pMesh->m_aNormalArray)
					{
						pMesh->m_aNormalArray[j].x   = n[0];
						pMesh->m_aNormalArray[j].y   = n[1];
						pMesh->m_aNormalArray[j].z   = n[2];
					}

					if(pMesh->m_aTexCoordArray)
					{
						pMesh->m_aTexCoordArray[j].u = t[0];
						pMesh->m_aTexCoordArray[j].v = t[1];
					}

					pMesh->m_aIndices[m] = ActualVertexArraySize++;

					goto end_elem;

					reuse_element:
						pMesh->m_aIndices[m] = l;

					end_elem:
						++m;
				}
			}


			pMesh->m_aVertexArray		= (ObjVertex*)  realloc(pMesh->m_aVertexArray,ActualVertexArraySize*sizeof(ObjVertex));
			assert( pMesh->m_aVertexArray );

			if(pMesh->m_aNormalArray)
			{
				pMesh->m_aNormalArray   = (ObjNormal*)  realloc(pMesh->m_aNormalArray,ActualVertexArraySize*sizeof(ObjNormal));
				assert(pMesh->m_aNormalArray);
			}

			if(pMesh->m_aTexCoordArray)
			{
				pMesh->m_aTexCoordArray = (ObjTexCoord*)realloc(pMesh->m_aTexCoordArray,ActualVertexArraySize*sizeof(ObjTexCoord));
				assert(pMesh->m_aTexCoordArray);
			}

			/*
			**	Using the number of faces variable as the count of all the vertices in the array
			*/
			pMesh->m_iNumberOfFaces*=3;

			/*
			**	Set the mode to vertex array
			*/
			pMesh->m_iMode = VERTEX_ARRAY;

			/*
			**	Zero the data counts
			*/
			pMesh->m_iNumberOfNormals	= 0;
			pMesh->m_iNumberOfTexCoords = 0;
			pMesh->m_iNumberOfVertices	= 0;

			/*
			**	Free the old data....
			*/
			if(pVT)
				free(pVT);
			if(pNT)
				free(pNT);
			if(pTT)
				free(pTT);
			if(pMesh->m_aFaces)
				free(pMesh->m_aFaces);
			pMesh->m_aFaces = NULL;

		}
		break;

	default:
		break;
	}
}
